#    Copyright 2020 Solomon M. Adams

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import configparser

from . import LOGGING

BASE_ACCESSION = {
    ".": "NA"
}

GRCH37_ACCESSIONS = {
                "NC_000001.10": "1",
                "NC_000002.11": "2",
                "NC_000003.11": "3",
                "NC_000004.11": "4",
                "NC_000005.9": "5",
                "NC_000006.11": "6",
                "NC_000007.13": "7",
                "NC_000008.10": "8",
                "NC_000009.11": "9",
                "NC_000010.10": "10",
                "NC_000011.9": "11",
                "NC_000012.11": "12",
                "NC_000013.10": "13",
                "NC_000014.8": "14",
                "NC_000015.9": "15",
                "NC_000016.9": "16",
                "NC_000017.10": "17",
                "NC_000018.9": "18",
                "NC_000019.9": "19",
                "NC_000020.10": "20",
                "NC_000021.8": "21",
                "NC_000022.10": "22",
                "NC_000023.10": "23",
                "NC_000024.9": "24"
            }

GRCH38_ACCESSIONS = {
                "NC_000001.11": "1",
                "NC_000002.12": "2",
                "NC_000003.12": "3",
                "NC_000004.12": "4",
                "NC_000005.10": "5",
                "NC_000006.12": "6",
                "NC_000007.14": "7",
                "NC_000008.11": "8",
                "NC_000009.12": "9",
                "NC_000010.11": "10",
                "NC_000011.10": "11",
                "NC_000012.12": "12",
                "NC_000013.11": "13",
                "NC_000014.9": "14",
                "NC_000015.10": "15",
                "NC_000016.10": "16",
                "NC_000017.11": "17",
                "NC_000018.10": "18",
                "NC_000019.10": "19",
                "NC_000020.11": "20",
                "NC_000021.9": "21",
                "NC_000022.11": "22",
                "NC_000023.11": "23",
                "NC_000024.10": "24"
            }

IUPAC_CODES = {
                'R': ['A', 'G'], 
                'Y': ['C', 'T'], 
                'S': ['G', 'C'], 
                'W': ['A', 'T'], 
                'K': ['G', 'T'], 
                'M': ['A', 'C'], 
                'N': ['A', 'C', 'T', 'G'],
                'A': ['A'],
                'C': ['C'],
                'T': ['T'],
                'G': ['G']                
            }

class ConfigData:

    def __init__(self, config_path: str = "config.ini") -> None:
        """
        Create a new config class that is used throughout to set various parameters

        Args:
            config_path (str): path to a configuration file - can be a "dummy" value. 
            write_config (bool, optional): Write a new configuration file to config_path?. Defaults to False.
        """
        self.config = configparser.ConfigParser()
        self.config.read(config_path)
        self._chromosome_accessions()
        self._iupac_codes()
        self._variant_query_params()
        self._missing_params()
        self._lp_params()
    
    def _chromosome_accessions(self) -> None:
        """
        Set chromosome accessions
        """
        try:
            self.CHROMOSOME_ACCESSIONS = {accession.upper(): chromosome for accession, chromosome in self.config["CHROMOSOME ACCESSIONS"].items()}
        except KeyError:
            LOGGING.info("Chromosome accessions defaulting to GRCh38")
            self.CHROMOSOME_ACCESSIONS = {**GRCH38_ACCESSIONS, **GRCH37_ACCESSIONS}
            self.config["CHROMOSOME ACCESSIONS"] = self.CHROMOSOME_ACCESSIONS
        self.CHROMOSOME_ACCESSIONS = {**self.CHROMOSOME_ACCESSIONS, **BASE_ACCESSION}
    
    def _iupac_codes(self) -> None:
        """
        Set IUPAC nucleotide codes
        """
        # Get IUPAC codes from config, if not found - use default and update
        try:
            self.IUPAC_CODES = {code.upper(): list(nucleotides) for code, nucleotides in self.config["IUPAC CODES"].items()}
        except KeyError:
            LOGGING.info("IUPAC Codes set to default values.")
            self.IUPAC_CODES = IUPAC_CODES
            self.config["IUPAC CODES"] = {code: "".join(nucleotides) for code, nucleotides in self.IUPAC_CODES.items()}
    
    def _variant_query_params(self) -> None:
        """
        Set variant query parameters
        """
        # Get variant query parameters, if not found - add defaults
        try:
            self.VARIANT_QUERY_PARAMETERS = self.config["VARIANT QUERY PARAMETERS"]
        except KeyError:
            LOGGING.info("Using a 1kb 5' and 3' offset.")
            self.VARIANT_QUERY_PARAMETERS = {
                "5p_offset": 1000,
                "3p_offset": 1000
            }
            self.config["VARIANT QUERY PARAMETERS"] = self.VARIANT_QUERY_PARAMETERS
    
    def _missing_params(self) -> None:
        try:
            self.MISSING_DATA_PARAMETERS = self.config["MISSING DATA PARAMETERS"]
        except KeyError:
            LOGGING.info("Missing variants will be assumed to match reference")
            self.MISSING_DATA_PARAMETERS = {
                "missing_variants": 99
            }
            self.config["MISSING DATA PARAMETERS"] = self.MISSING_DATA_PARAMETERS
    
    def _lp_params(self) -> None:
        try:
            self.LP_PARAMS = self.config["LINEAR PROGRAM PARAMETERS"]
        except KeyError:
            self.LP_PARAMS = {
                "optimal_decay": 0,
                "max_haps": 2
            }
            self.config["LINEAR PROGRAM PARAMETERS"] = self.LP_PARAMS
    
    def write_config(self, config_path: str) -> None:
        """
        Write default/modified parameters to file at config_path

        Args:
            config_path (str): path to configuration file
        """
        with open(config_path, "w") as configfile:	
            self.config.write(configfile)
    


#    Copyright 2020 Solomon M. Adams

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import csv

import numpy as np

from pysam import VariantFile

from .template import PATH

from . import LOGGING


class VarFile:
    def __init__(self, vcf_file: str, sample: str = None, vcf_file_index: str = None, config = None) -> None:
        """
        VarFile object, basically a wrapper for pysam VariantFile
        
        Args:
            vcf_file (str): path to VCF/VCF.GZ/BCF file (needs to be indexed)
        """
        self.vcf_file = VariantFile(vcf_file, index_filename = vcf_file_index)
        if sample:
            self.samples = [sample]
        else:
            self.samples = list(self.vcf_file.header.samples)
        self.vcf_file.subset_samples(self.samples)
    
    def _get_alleles(self, sample, var_type):
        """
        Get alleles if coded as GT, catch if coded as CN (for CNV)
        Will always assume that at least one allele is one copy (unless CN == 0)
        Not necessarily accurate, but plain count data is inherently unphased so the distiction
        between (for example) 1/3 vs 2/2 is arbitrary. 
        """
        alleles = sample.alleles
        if len(alleles) == 0 and var_type == "CNV":
            try:
                cn = sample["CN"]
                if cn == 0:
                    alleles = ("0", "0")
                else:
                    alleles = ("1", str(cn - 1)) #TODO this in untested
            except KeyError:
                alleles = None # This is probably going to cause an issue eventually
        return alleles


    def get_range(self, chrom: str, minloc: int, maxloc: int) -> dict:
        """
        Returns a range of variants for all samples in a VCF file
        
        Args:
            chrom (str): chromosome
            minloc (int): starting position
            maxloc (int): ending position
        
        Returns:
            dict: variants with a common ID schema that is matched by other methods
        """
        positions_out = {}
        try:
            positions = self.vcf_file.fetch(str(chrom), minloc, maxloc)
        except ValueError:
            positions = self.vcf_file.fetch(f"chr{chrom}", minloc, maxloc)
        for position in positions:
            chrom = position.chrom.strip("chr")
            var_type = "SID"
            try:
                var_type = position.info["SVTYPE"]
            except KeyError:
                pass
            positions_out[f"c{chrom}_{position.pos}_{var_type}"] = {
                sample: {
                    "alleles": self._get_alleles(position.samples[sample], var_type), "phased": position.samples[sample].phased, "ref": position.ref} for sample in self.samples}
        return positions_out
        
def get_alleles(gene: object, subjects: list) -> list:
    """
    Prep for the ref/alt columns in a VCF

    Args:
        gene (Gene): gene object
        subjects (list): list of subjects

    Returns:
        list: list of all possible alt alleles
    """
    ref = gene.reference
    alts = []
    for s in subjects:
        subject_haps = [h for h in s.called_haplotypes[str(gene)]["HAPS"][0]]
        alts += np.array(subject_haps).flatten().tolist()
    try:
        alts = list(filter((ref).__ne__, alts))
    except ValueError:
        pass
    if len(alts) == 0:
        alts = ["NON_REF"]
    return([ref] + list(set(alts)))

def get_dosage(haps: list, alleles: list) -> list:
    """
    Get index for GT fiels for a given sample

    Args:
        haps (list): possible haplotypes
        alleles (list): alleles

    Returns:
        list: sample dosages for each allele
    """
    return [alleles.index(s) for s in haps]

def get_samples(gene_name: str, subjects: list, alleles: list) -> list:
    """
    Generate the sample/format field for each sample in the multi-sample output VCF

    Args:
        gene_name (str): gene name
        subjects (list): list of subjects to be written
        alleles (list): subject alleles

    Returns:
        list: list of format field for each each sample
    """
    formats = []
    for s in subjects:
        calls = s.called_haplotypes[gene_name]["HAPS"]
        formats.append(
            {
                "GT": get_dosage(calls[0][0], alleles),
                "VA": calls[1][0],
                "HC": 1 / len(calls[0])
            }
        )
    return formats


def write_variant_file(directory: str, subjects: [], prefix: str, genes: list) -> None:
    """
    Write the output VCF file

    Args:
        directory (str): output directory
        subjects ([type]): list of subjects
        prefix (str): prefix for filename
        genes (list): list of gene objects
    """
    contigs = list(set([f"chr{gene.chromosome.strip('chr')}" for gene in genes]))
    template = VariantFile(PATH + "/template.vcf", "r")
    outfile = VariantFile(directory+f"/{prefix}.haplotypes.vcf", "w", header = template.header)
    for contig in contigs:
        outfile.header.add_line(f"##contig=<ID={contig},length=0>")
    for sub in subjects:
        outfile.header.add_sample(str(sub))
    for gene in genes:
        alleles = get_alleles(gene, subjects)
        nr =outfile.new_record(
            contig = f"chr{gene.chromosome}",
            start = gene.min,
            stop = gene.max,
            alleles = [f'<{a.replace(str(gene), "").replace("(star)", "*")}>' for a in alleles],
            id = f"{str(gene)}_pgx",
            qual = None,
            filter = None,
            info = {"VARTYPE": "HAP"},
            samples = get_samples(str(gene), subjects, alleles)
        )
        outfile.write(nr)
    outfile.close()

def write_flat_file(directory: str, subjects: [], prefix: str) -> None:
    """
    Writes the output flat file

    Args:
        directory (str): output directory
        subjects ([type]): list of subjects
        prefix (str): prefix for filename
    """
    lines = []
    for subject in subjects:
        for gene, haps in subject.called_haplotypes.items():
            for i in range(len(haps["HAPS"][0])):
                lines.append({
                    "SUBJECT": str(subject),
                    "GENE": gene,
                    "GENOTYPE": "/".join(haps["HAPS"][0][i]),
                    "VARIANTS": "|".join(haps["HAPS"][1][i]),
                    "CONFIDENCE": 1 / len(haps["HAPS"][0])
                })
    with open(directory + f"/{prefix}.haplotypes.tsv", "w") as flat_out:
        flat_file = csv.DictWriter(flat_out, ["SUBJECT", "GENE", "GENOTYPE", "VARIANTS", "CONFIDENCE"], delimiter = "\t")
        flat_file.writeheader()
        flat_file.writerows(lines)



#    Copyright 2020 Solomon M. Adams

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import sys
import itertools

import pandas as pd
import numpy as np

from pulp import *
from .gene import AbstractGene
from . import LOGGING

class NoVariantsException(Exception):
    """
    Exception to call if a sample is attempted that has zero variants defined. 
    """
    pass

class Haplotype:

    def __init__(self, gene: AbstractGene, sample_prefix: str, config = None) -> None:
        """
        Create a new haplotype object
        This object is not a subclass, but inherits data from the Gene class
        Conceptually, Gene is a fairly abstract class that has meta information used
        by the subject and haplotype classes. 

        Args:
            gene (Gene): gene.Gene object
            sample_prefix (str): Sample ID 
        """
        self.phased = gene.phased
        self.config = config
        self.solver = gene.solver
        self.matched = False
        self.sample_prefix = sample_prefix
        self.genotypes = gene.get_sample_vars(sample_prefix)
        if len(self.genotypes) == 0:
            raise NoVariantsException
        self.translation_table = gene.get_translation_table_copy()
        self.chromosome = gene.chromosome
        self.version = gene.version
        self.reference = gene.reference
    
    def table_matcher(self) -> None:
        """
        Matches variants in the translation table with the subject's variants
        """
        self.matched = True
        matches = self.translation_table.apply(self._match, axis = 1)
        self.translation_table["MATCH"] = [m[0] for m in matches]
        self.translation_table["STRAND"] = [m[1] for m in matches]
        self.translation_table["VAR_ID"] = self.translation_table.apply(
                lambda x: f'{x["ID"]}_{str(x.iloc[7]).strip("<>")}',
                axis = 1
                )
        self.translation_table = self.translation_table.drop(self.translation_table.index[self.translation_table["MATCH"] == 99].tolist())
        haps = self.translation_table["Haplotype Name"]
        no_match = self.translation_table[self.translation_table["MATCH"] == 0].iloc[:,0] # Haplotypes where there is any variant not matching
        drops = []
        for i in no_match.unique():
            if sum([i == k for k in no_match]) > 0:
                drops.append(i)
        self.translation_table = self.translation_table[~self.translation_table.iloc[:,0].isin(drops)] # Drop haplotypes that don't match 100%
        self.variants = self.translation_table.loc[:,["VAR_ID", "MATCH", "STRAND", "Type", "Variant Start"]].drop_duplicates() # List of matched variants
        self.haplotypes = [hap for hap in self.translation_table.iloc[:,0].unique().tolist()] # List of possible haplotypes

    def _mod_vcf_record(self, alt: str, ref: str) -> str:
        """
        Modifies record from VCF to standardized form
        
        Args:
            alt (str): alt allele
            ref (str): ref allele
        
        Returns:
            str: reformatted alt allele
        """
        if alt is None:
            return "-"
        if "<" in alt:
            return f"s{alt.strip('<>')}"
        elif len(ref) > len(alt):
            return "id-"
        elif len(ref) < len(alt):
            return f'id{alt[1:]}' # Remove first position
        else:
            return f's{alt}'

    def _mod_tt_record(self, var_type: str, alt: str) -> list:
        """
        Modifies the translation table ref to a standardized form
        
        Args:
            var_type (str): insertion, deletion, or substitution
            alt (str): allele from translation table
        
        Returns:
            [list]: modified allele as list based on iupac
        """
        alt = alt.strip("<>")
        if var_type == "insertion":
            return [f'id{alt}']
        elif var_type == "deletion":
            return [f'id-']
        else:
            try:
                return [f's{a}' for a in self.config.IUPAC_CODES[alt]]
            except KeyError:
                return [f's{alt}']

    def _match(self, row: pd.core.series.Series) -> (int, int):
        """
        Evaluate match in a single translation table row with a sample

        Args:
            row (pd.core.series.Series): single row from translation table
            genotypes ([type]): list of genotypes

        Returns:
            int: 99 (missing), 0, 1, or 2 (corresponds to the number of matched alleles for a particular position)
        """
        strand = 0
        if row.iloc[8] in ["insertion", "deletion"]:
            new_pos = int(row["ID"].split("_")[1]) - 1
            ID = f'{row["ID"].split("_")[0]}_{new_pos}_SID'
        else:
            ID = row["ID"]
        try:
            genotype = self.genotypes[ID]
        except KeyError: # Not in VCF
            return int(self.config.MISSING_DATA_PARAMETERS["missing_variants"]), strand
        try:
            vcf_geno = [self._mod_vcf_record(g, genotype["ref"]) for g in genotype["alleles"]]
        except AttributeError:
            return int(self.config.MISSING_DATA_PARAMETERS["missing_variants"]), strand
        if vcf_geno == ["-", "-"]:
            return int(self.config.MISSING_DATA_PARAMETERS["missing_variants"]), strand
        tt_alt_geno = self._mod_tt_record(row.iloc[8], row.iloc[7])
        alt_matches = sum([vcf_geno.count(a) for a in tt_alt_geno])
        if alt_matches == 1 and genotype["phased"]:
            strand = 1 if max([vcf_geno.index(a) for a in tt_alt_geno]) == 1 else -1
        elif alt_matches == 2 and genotype["phased"]:
            strand = 3
        return alt_matches, strand
    
    def _haps_from_prob(self, lp_problem: object) -> tuple:
        """
        Take a optimally solved lp problem
        Produce called haplotypes

        Args:
            lp_problem (object): solved lp problem

        Returns:
            tuple: called haplotypes and associated information
        """
        is_ref = False
        haps = []
        variants = []
        for v in lp_problem.variables():
            if v.varValue:
                if v.varValue > 0:
                    if v.name.split("_")[0] == f'c{self.chromosome}':
                        variants.append(v.name)
                    else:
                        haps.append((v.name, v.varValue))
        if len(haps) == 0:
            called = [self.reference, self.reference]
            is_ref = True
        elif len(haps) == 2:
            called = [haps[0][0], haps[1][0]]
        else:
            called = np.array([np.repeat(i[0], i[1]) for i in haps]).flatten().tolist()
            if len(called) == 1:
                called.append(self.reference)
        return called, variants, len(haps), is_ref
    
    def _solve(self, hap_prob: object) -> object:
        if self.solver == "GLPK":
            hap_prob.solve(GLPK(msg=0))
        else:
            hap_prob.solve()

    
    def lp_hap(self) -> tuple:
        """
        Build and run the LP problem

        Returns:
            tuple: list of possible haplotypes and list of associated variants
        """
        possible_haplotypes = []
        haplotype_variants = []
        num_vars = self.variants.shape[0]
        num_haps = len(self.haplotypes)
        hap_vars = []
        for hap in self.haplotypes:
            trans = self.translation_table[self.translation_table.iloc[:,0] == hap]
            hap_vars.append([1 if var in trans["VAR_ID"].unique() else 0 for var in self.variants["VAR_ID"]])

        hap_prob = LpProblem("Haplotype Optimization", LpMaximize)

        # Define the haplotypes and variants variables
        haplotypes = [LpVariable(hap, cat = "Integer", lowBound=0, upBound=2) for hap in self.haplotypes]
        variants = [LpVariable(var, cat = "Binary") for var in self.variants["VAR_ID"]]
        # Set constraint of two haplotypes selected
        hap_prob += (lpSum(haplotypes[i] for i in range(num_haps)) <= int(self.config.LP_PARAMS["max_haps"])) # Cannot choose more than x haplotypes (may be increased to find novel sub-alleles)
       # Limit alleles that can be chosen based on zygosity
        for i in range(num_vars): # Iterate over every variant
            # A variant allele can only be used once per haplotype, up to two alleles per variant
            hap_prob += (variants[i] <= (lpSum(hap_vars[k][i] * haplotypes[k] for k in range(num_haps))))
            # A given variant cannot be used more than "MATCH"
            hap_prob += ((lpSum(hap_vars[k][i] * haplotypes[k] for k in range(num_haps))) <= self.variants.iloc[i,1] * variants[i])
            # Any CNV variants defined, if matched with a haplotype, MUST be used
            # Otherwise, variants like CYP2D6*5 will be missed by the other methods
            if self.variants.iloc[i,3] == "CNV":
                hap_prob += ((lpSum(hap_vars[k][i] * haplotypes[k] for k in range(num_haps))) == self.variants.iloc[i,1])
        if self.phased:
            for i in range(num_haps):
                hap_prob += lpSum(haplotypes[i] * self._get_strand_constraint(i, []).size) <= 1 # max one strand
            hap_prob += lpSum(haplotypes[i] * self._get_strand_constraint(i, [0])[0] for i in range(num_haps)) <= 1
            hap_prob += lpSum(haplotypes[i] * self._get_strand_constraint(i, [0])[0] for i in range(num_haps)) >= -1
        # Set to maximize the number of variant alleles used
        hap_prob += lpSum(
            self.translation_table[
                    (self.translation_table.iloc[:,0] == self.haplotypes[i]) &
                    (self.translation_table["MATCH"] > 0)
                ].shape[0] * haplotypes[i] for i in range(num_haps))
        self._solve(hap_prob)
        if hap_prob.status != 1:
            if self.phased:
                LOGGING.warning(f"No feasible solution found, {self.sample_prefix} will be re-attempted with phasing off.")
                self.phased = False
                return None, None
            else:
                LOGGING.warning(f"No feasible solution found, {self.sample_prefix} will not be called")
                return [], []
        else:
            called, variants, hap_len, is_ref = self._haps_from_prob(hap_prob)
            if is_ref:
                possible_haplotypes.append(tuple(called))
                haplotype_variants.append(tuple(variants))
                return possible_haplotypes, haplotype_variants
            max_opt = hap_prob.objective.value()
            opt = max_opt
            while opt >= (max_opt - float(self.config.LP_PARAMS["optimal_decay"])) and not is_ref and hap_prob.status >= 0:
                possible_haplotypes.append(tuple(sorted(called)))
                haplotype_variants.append(tuple(sorted(variants)))
                hap_prob += lpSum([h.value() * h for h in haplotypes]) <= hap_len - 1
                self._solve(hap_prob)
                if hap_prob.status != 1:
                    break
                opt = hap_prob.objective.value()
                new_called, variants, hap_len, is_ref = self._haps_from_prob(hap_prob)
                if new_called == called or len(new_called) == 0:
                    break
                called = new_called

            return possible_haplotypes, haplotype_variants
    

    def _get_strand_constraint(self, i: int, default: list) -> list:
        """
        Helps to assemble the constraint for phased data
        Looks at all strands that are part of a haplotype
        Removes homozygous calls
        Args:
            i (int): haplotype index
            default (list): default return if nothing matches or if all are homozygous

        Returns:
            list: [description]
        """
        sc = self.translation_table[self.translation_table.iloc[:,0] == self.haplotypes[i]]["STRAND"].unique()
        sc = np.delete(sc, np.where(sc == [3]))
        return(sc if sc.size > 0 else np.array(default))


    def optimize_hap(self) -> ():
        """
        Solve for the most likely diplotype

        Returns:
            (): Results
        """
        if not self.matched:
            print("You need to run the table_matcher function with genotyped before you can optimize")
            sys.exit(1)
        called, variants = self.lp_hap()
        if called is None:
            # Happens when a phased call attempt fails
            called, variants = self.lp_hap()
        if len(called) > 1:
            LOGGING.warning(f"Multiple genotypes possible for {self.sample_prefix}.")
        return called, variants
      



#    Copyright 2020 Solomon M. Adams

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import pandas as pd
import sys

from . import LOGGING
from .vcf import VarFile

class AbstractGene:
    """
    Abstract gene class, conains top level information that is available to muliple
    names haplotypes and subjects. 
    """

    def __init__(self, translation_table: str, vcf: VarFile = None, 
                    variants = None, solver: str = "CBC",
                    config = None, phased = False) -> None:
        """
        Create a Gene object
        
        Args:
            translation_table (str): path to translation table
            vcf (VarFile): parsed VCF object from vcf.VarFile
        """
        self.config = config
        self.phased = phased
        self.solver = solver
        self.gene = None
        self.accession = None
        self.read_translation_table(translation_table)
        self.gene = self.translation_table.iloc[-1, 1]
        self.max = self.translation_table.iloc[:,5].dropna().max() + int(self.config.VARIANT_QUERY_PARAMETERS["5p_offset"])
        self.min = self.translation_table.iloc[:,4].dropna().min() - int(self.config.VARIANT_QUERY_PARAMETERS["3p_offset"])
        if vcf:
            self.variants = vcf.get_range(self.chromosome, self.min, self.max)
        else:
            self.variants = variants

    def __str__(self):
        return self.gene

    def __repr__(self):
        return self.gene
    
    def get_sample_vars(self, sample: str) -> dict:
        """
        The gene contains variants for all samples in the VCF
        This function parses variants for a single sample. 

        Args:
            sample (str): sample ID

        Returns:
            dict: single sample variants from VCF
        """
        sample_vars = {}
        for var_id, sub_vars in self.variants.items():
            try:
                sample_vars[var_id] = sub_vars[sample]
            except KeyError:
                pass
        return sample_vars
    
    def get_translation_table_copy(self) -> pd.DataFrame:
        """
        Get a deep copy of a translation table that can be modified
        for a single subject

        Returns:
            pd.DataFrame: deep copy of the associated translation table
        """
        return(self.translation_table.copy(deep = True))
    
    def _merge_tables(self, translation_table, cnv_table):
        new_rows = []
        cnv_table["BASE"] = cnv_table.apply(lambda x: x["Haplotype Name"].split("_")[0], axis = 1)
        cnv_table["SUFFIX"] = cnv_table.apply(lambda x: x["Haplotype Name"].split("_")[-1], axis = 1)
        translation_table["BASE"] = translation_table.apply(lambda x: x["Haplotype Name"].split(".")[0], axis = 1)
        translation_table["SUFFIX"] = translation_table.apply(lambda x: x["Haplotype Name"].split(".")[-1], axis = 1)
        for index, row in cnv_table.iterrows():
            trans_base = translation_table[translation_table["BASE"] == row["BASE"]]
            trans_suffixes = trans_base["SUFFIX"]
            for suf in trans_suffixes:
                new_rows.append(
                    {
                        "Haplotype Name": f'{row["BASE"]}.{suf}_{row["SUFFIX"]}',
                        "Gene": row["Gene"],
                        "rsID": row["rsID"],
                        "ReferenceSequence": row["ReferenceSequence"],
                        "Variant Start": row["Variant Start"],
                        "Variant Stop": row["Variant Stop"],
                        "Reference Allele": row["Reference Allele"],
                        "Variant Allele": row["Variant Allele"],
                        "Type": row["Type"]
                    }
                )
            for i, r in trans_base.iterrows():
                new_rows.append(
                    {
                        "Haplotype Name": f'{row["BASE"]}.{r["SUFFIX"]}_{row["SUFFIX"]}',
                        "Gene": r["Gene"],
                        "rsID": r["rsID"],
                        "ReferenceSequence": r["ReferenceSequence"],
                        "Variant Start": r["Variant Start"],
                        "Variant Stop": r["Variant Stop"],
                        "Reference Allele": r["Reference Allele"],
                        "Variant Allele": r["Variant Allele"],
                        "Type": r["Type"]
                    }
                )
        new_table = translation_table.drop(["BASE", "SUFFIX"], axis = 1)
        added_rows = pd.DataFrame(new_rows)
        return pd.concat([new_table, added_rows, cnv_table], ignore_index=True)

    def get_type(self, vtype: str) -> str:
        """
        Simple helper function to assign a CNV type if a variant is not SID

        Args:
            vtype (str): SID or CNV

        Returns:
            str: CNV if CNV, else SID
        """
        return "CNV" if vtype == "CNV" else "SID"

    def read_translation_table(self, translation_table: str) -> pd.DataFrame:
        """Read and process a translation table
        
        Args:
            translation_table (str): path to translation table file
            config (ConfigData): config.Config object
        
        Returns:
            pd.DataFrame: translation table as pandas data frame
        """
        with open(translation_table, 'rt') as trans_file:
            self.version = trans_file.readline().strip("#version=\n\t")
        self.translation_table = pd.read_csv(
                                        translation_table, 
                                        skipinitialspace = True,
                                        delim_whitespace=True,
                                        skiprows = 2,
                                        na_values= {4: ".", 5: "."},
                                        dtype = {4: pd.Int64Dtype(), 5: pd.Int64Dtype()},
                                        names = ["Haplotype Name", "Gene", 
                                                "rsID", "ReferenceSequence",
                                                "Variant Start", "Variant Stop",
                                                "Reference Allele", "Variant Allele",
                                                "Type"]
        )
        try:
            cnv_table = pd.read_csv(
                                    translation_table.replace(".tsv", ".cnv"),
                                    skipinitialspace = True,
                                    delim_whitespace=True,
                                    skiprows = 2,
                                    na_values= {4: ".", 5: "."},
                                    dtype = {4: pd.Int64Dtype(), 5: pd.Int64Dtype()},
                                    names = ["Haplotype Name", "Gene", 
                                            "rsID", "ReferenceSequence",
                                            "Variant Start", "Variant Stop",
                                            "Reference Allele", "Variant Allele",
                                            "Type"]
            )
        except FileNotFoundError:
            LOGGING.info("No CNV definition file found, proceeding with only SID variants.")
            cnv_table = None
        if cnv_table is not None:
            self.translation_table = self._merge_tables(self.translation_table, cnv_table)
        self.translation_table.iloc[:,0] = self.translation_table.apply(lambda x: x.iloc[0].replace("*", "(star)"), axis = 1)
        try:
            self.reference = self.translation_table[self.translation_table["rsID"] == "REFERENCE"]["Haplotype Name"][0]
        except IndexError:
            self.reference = "REF"
        self.translation_table = self.translation_table[self.translation_table["ReferenceSequence"] != "."]
        self.accession = self.translation_table.iloc[-1, 3]
        self.chromosome = self.config.CHROMOSOME_ACCESSIONS[self.accession]
        self.translation_table["ID"] = self.translation_table.apply(lambda x: f"c{self.chromosome}_{x['Variant Start']}_{self.get_type(x['Type'])}", axis = 1)
        self.translation_table["EXCLUDE"] = 0
        


#    Copyright 2020 Solomon M. Adams

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import sys

from .haplotype import Haplotype, NoVariantsException
from .gene import AbstractGene
from . import LOGGING

class Subject:

    def __init__(self, prefix: str, genes: [AbstractGene], config = None) -> None:
        """
        Subject object - manages data and functions for a single sample in a VCF file
        
        Args:
            prefix (str): Subject ID (comes from the VCF file)
            genes ([Gene]): List of gene.Gene objects
        """
        self.config = config
        self.prefix = prefix
        self.called_haplotypes = {}
        for gene in genes:
            try:
                haplotype = Haplotype(gene, self.prefix, self.config)
                haplotype.table_matcher()
                self.called_haplotypes[str(gene)] = {
                    "HAPS": haplotype.optimize_hap(),
                    "CONTIG": gene.chromosome}
            except NoVariantsException:
                LOGGING.warning(f"{self.prefix} has no variants, returning NA")
                self.called_haplotypes[str(gene)] = {
                    "HAPS": ("NA", "NA", "NA", "NA"),
                    "CONTIG": gene.chromosome}

    
    def __str__(self):
        return self.prefix

    def __repr__(self):
        return self.prefix
    





#    Copyright 2020 Solomon M. Adams

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from .vcf import VarFile
from .gene import AbstractGene
from .haplotype import Haplotype
from .subject import Subject
from . import get_config

def get_haps_from_vcf(translation_table_path: str, vcf_file_path: str, 
                        sample_id: str, solver: str = "CBC", 
                        config_path: str = None, phased = False) -> tuple:
    """
    Provide a VCF file, sample ID, and translation table
    Get called haplotypes and additional information

    Args:
        translation_table_path (str): [description]
        vcf_file_path (str): [description]
        sample_id (str): [description]
        config_path ([type], optional): [description]. Defaults to None.

    Returns:
        tuple: translation_table_version, called_haplotypes, variants_associated_with_haplotye, matched_translation_table
    """
    config = get_config(config_path)
    vcf = VarFile(vcf_file_path, sample_id)
    gene = AbstractGene(translation_table_path, vcf = vcf, solver = solver, config = config, phased = phased)
    haplotype = Haplotype(gene, sample_id, config = config)
    haplotype.table_matcher()
    return haplotype.optimize_hap()

def get_haps_from_variants(translation_table_path: str, vcf_data: str, 
                            sample_id: str, solver: str = "CBC", 
                            config_path: str = None, phased = False) -> tuple:
    """
    Same as get_haps_from_vcf, but bypasses the VCF file so that you can provide formatted variants from another input
    Get called haplotypes and additional information

    Args:
        translation_table_path (str): [description]
        vcf_file_path (str): [description]
        sample_id (str): [description]
        config_path ([type], optional): [description]. Defaults to None.

    Returns:
        tuple: translation_table_version, called_haplotypes, variants_associated_with_haplotye, matched_translation_table
    """
    config = get_config(config_path)
    gene = AbstractGene(translation_table_path, variants = vcf_data, solver = solver, config = config, phased = phased)
    haplotype = Haplotype(gene, sample_id, config = config)
    haplotype.table_matcher()
    return haplotype.optimize_hap()



#    Copyright 2020 Solomon M. Adams

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import yaml
import glob
import os
from parameterized import parameterized

from hiMoon import gene, vcf, subject, himoon, get_config

PATH = os.path.dirname(os.path.abspath(__file__))

CONFIG = get_config()

def load_definition_file(file_path: str) -> dict:
    """
    Wrapper for yaml.load

    Args:
        file_path (str): path/to/yml/file

    Returns:
        dict: dict representation of yml
    """
    with open(file_path, "r") as sample_allele_data:
        try:
            return(yaml.load(sample_allele_data, Loader = yaml.CLoader))
        except AttributeError: # Happens if libyaml headers are not available
            return(yaml.load(sample_allele_data, Loader = yaml.Loader))

def prep_samples() -> []:
    gene_samples = []
    definition_files = glob.glob(PATH + "/test_files/*.yaml")
    for definition_file in definition_files:
        sample_alleles = load_definition_file(definition_file)
        vcf_file = vcf.VarFile(PATH + "/test_files/" + sample_alleles["VCF"])
        gene_obj = gene.AbstractGene(PATH + "/test_files/" + sample_alleles["TRANSLATION_TABLE"], vcf = vcf_file, config = CONFIG)
        gene_samples += [(subject.Subject(sample["ID"], genes = [gene_obj], config = CONFIG), gene_obj, sample) for sample in sample_alleles["SAMPLES"]]
    return(gene_samples)

def rm_sub_allele(allele: str) -> str:
    """
    Remove suballele from a called named allele.
    (star)2.001 -> (star)2
    (star)2.001_x2 -> (star)2x2

    Args:
        allele (str): [description]

    Returns:
        str: [description]
    """
    sv = allele.split("_")[-1] if "_" in allele else None
    main_allele = allele.split(".")[0]
    return "".join([main_allele, sv]) if sv else main_allele

@parameterized.expand(prep_samples())
def test_sample_haplotypes(subj, gene_obj, sample):
    possible_haplotypes = []
    haps = subj.called_haplotypes[str(gene_obj)]["HAPS"][0]
    for hap in haps:
        possible_haplotypes.append(sorted([rm_sub_allele(a) for a in hap]))
    print(f"{sample['ID']} @ {str(gene_obj)} should be {'/'.join(sample['ALLELES'])}. Found {possible_haplotypes}")
    assert sorted(sample["ALLELES"]) in possible_haplotypes
    